/*-
 * Copyright (c) 2012 Brendan Fabeny
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

/*
 * Stopgap replacements for missing standard libm functions.  Slower than many 
 * other libm functions, but more accurate than many naive replacements. No
 * exception handling, and no use of the ternary values (yet).
 */

#include <sys/param.h>

#include <fenv.h>
#pragma STDC FENV_ACCESS ON
#include <complex.h>

#if 0
#define MPFR_USE_FILE
#define MPFR_USE_VA_LIST
#endif
#include "mpc.h"

mpc_rnd_t
getrnd_mpc(void)
{
	mpc_rnd_t rnd;
	int rndmode;

	rndmode=fegetround();

	switch (rndmode) {
	case FE_TONEAREST:
		rnd = MPC_RNDNN;
	case FE_DOWNWARD:
		rnd = MPC_RNDDD;
	case FE_UPWARD:
		rnd = MPC_RNDUU;
	case FE_TOWARDZERO:
		rnd = MPC_RNDZZ;
	}

	return (rnd);
}

mpfr_rnd_t
getrnd_mpfr(void)
{
	mpfr_rnd_t rnd;
	int rndmode;
	
	rndmode=fegetround();

	switch (rndmode) {
	case FE_TONEAREST:
		rnd = MPFR_RNDN;
	case FE_DOWNWARD:
		rnd = MPFR_RNDD;
	case FE_UPWARD:
		rnd = MPFR_RNDU;
	case FE_TOWARDZERO:
		rnd = MPFR_RNDZ;
	}

	return (rnd);
}

mpfr_rnd_t
getrnd_mpfr_from_mpc(mpc_rnd_t rnd)
{
	mpfr_rnd_t rndmpfr;

	switch (rnd) {
	case MPC_RNDNN:
		rndmpfr = MPFR_RNDN;
	case MPC_RNDDD:
		rndmpfr = MPFR_RNDD;
	case MPC_RNDUU:
		rndmpfr = MPFR_RNDU;
	case MPC_RNDZZ:
		rndmpfr = MPFR_RNDZ;
	}

	return (rndmpfr);
}




#define	prec_d	64
#if defined(__arm__) || defined(__mips__) || defined(__powerpc__)
#define	prec_ld	64
#elif defined(__sparc__)
#define	prec_ld	128
#else
#define	prec_ld	80
#endif

/*
 * mpc_set_fc and mpc_get_fc are not in the API. Crude versions using casts to
 * double complex and back can be used, if prec_f is set equal to prec_d. To avoid
 * explicit excess intermediate precision for floats under the existing API
 * requires a few more steps, and linking to libm. It is probably not worth
 * the effort.
 */

#if 0
#define	prec_f	32

int
mpc_set_fc(mpc_t mx, float complex x, mpc_rnd_t rnd)
{
	float rx, ix;
	int i, j, k;
	mpfr_t mrx, mix;
	mpfr_rnd_t rnd2;
	
	rnd2=getrnd_mpfr_from_mpc(rnd);
	rx=crealf(x);
	ix=cimagf(x);
	mpfr_inits2(prec_f, mrx, mix, (mpfr_ptr) 0);
	i=mpfr_set_flt(mrx, rx, rnd2);
	j=mpfr_set_flt(mix, ix, rnd2);
	k=mpc_set_fr_fr(mx, mrx, mix, rnd);
	mpfr_clears(mrx, mix, (mpfr_ptr) 0);
	return (k);
}

float complex
mpc_get_fc (mpc_t mx, mpc_rnd_t rnd)
{
	int j, k;
	float rx, ix;
	mpfr_t mrx, mix;
	mpfr_rnd_t rnd2;
	
	rnd2=getrnd_mpfr_from_mpc(rnd);
	mpfr_inits2(prec_f, mrx, mix, (mpfr_ptr) 0);
	j=mpc_real(mrx, mx, rnd2);
	k=mpc_imag(mix, mx, rnd2);
	rx=mpfr_get_flt(mrx, rnd2);
	ix=mpfr_get_flt(mrx, rnd2);
	mpfr_clears(mrx, mix, (mpfr_ptr) 0);
	return (rx+ix*I);
}

#else
#define	prec_f	64
#define	prec_flt prec_f
#define	mpc_set_fc(mx, x, rnd)	mpc_set_dc(mx, (double complex)x, rnd)
#define	mpc_get_fc(mx, rnd)	(float complex)mpc_get_dc(mx,rnd)
#endif

#define	init_mpc(typesuffix, mx) \
	mpc_init3(mx, prec_ ## typesuffix, prec_ ## typesuffix)

#define	assign_mpc(typesuffix, mx, x, rnd) \
	mpc_set_ ## typesuffix ## c(mx, x, rnd)

#define	convert_mpc(typesuffix, mx, rnd) \
	mpc_get_ ## typesuffix ## c(mx,rnd)

#define	cleanup_mpc

#define	init_mpfr(typesuffix, mx) \
	mpfr_init2(mx, prec_ ## typesuffix)

#define	assign_mpfr(typesuffix, mx, x, rnd) \
	mpfr_set_ ## typesuffix(mx, x, rnd)

#define	convert_mpfr(typesuffix, mx, rnd) \
	mpfr_get_ ## typesuffix(mx,rnd)

#define	cleanup_mpfr	mpfr_free_cache ()

#define	FUNCTION1(libmfunc, libmtype1, libmtype2, typesuffix1, auxlib1, \
	typesuffix2, auxlib2, auxfunc) \
	libmtype2 \
	libmfunc(libmtype1 x) \
	{ \
		int i, j; \
		libmtype2 y; \
		auxlib1 ## _rnd_t rnd1; \
		auxlib1 ## _t mx; \
		auxlib2 ## _rnd_t rnd2; \
		auxlib2 ## _t my; \
		\
		rnd1=getrnd_ ## auxlib1(); \
		if (#auxlib1==#auxlib2) \
			rnd2=rnd1; \
		else \
			rnd2=MPC_RND_RE(rnd1); \
		init_ ## auxlib1(typesuffix1, mx); \
		init_ ## auxlib2(typesuffix1, my); \
		i=assign_ ## auxlib1(typesuffix1, mx, x, rnd1); \
		j=auxlib1 ## _ ## auxfunc(my, mx, rnd2); \
		y=convert_ ## auxlib2(typesuffix2, my, rnd2); \
		auxlib1 ## _clear(mx); \
		auxlib2 ## _clear(my); \
		cleanup_ ## auxlib1; \
		if (#auxlib1!=#auxlib2) \
			cleanup_ ## auxlib2; \
		\
		return (y); \
	}

#define	FUNCTION2(libmfunc, libmtype, typesuffix, auxlib, auxfunc) \
	libmtype \
	libmfunc(libmtype x, libmtype y) \
	{ \
		int i, j, k; \
		libmtype z; \
		auxlib ## _rnd_t rnd; \
		auxlib ## _t mx, my; \
		\
		rnd=getrnd_ ## auxlib(); \
		init_ ## auxlib(typesuffix, mx); \
		i=assign_ ## auxlib(typesuffix, mx, x, rnd); \
		init_ ## auxlib(typesuffix, my); \
		j=assign_ ## auxlib(typesuffix, my, y, rnd); \
		k=auxlib ## _ ## auxfunc(mx, mx, my, rnd); \
		z=convert_ ## auxlib(typesuffix, mx, rnd); \
		auxlib ## _clear(mx); \
		auxlib ## _clear(my); \
		cleanup_ ## auxlib; \
		\
		return (z); \
	}

#if __FreeBSD_version < 704102 || \
(__FreeBSD_version >= 800000 && __FreeBSD_version < 802502) || \
(__FreeBSD_version >= 900000 && __FreeBSD_version < 900027)

FUNCTION1(log2, double, double, d, mpfr, d, mpfr, log2)
FUNCTION1(log2f, float, float, flt, mpfr, flt, mpfr, log2)

#endif
#if __FreeBSD_version < 800007

#if 0
/* nan* will probably require explicit definitions */
nan
nanf
nanl
/* Need a new macro or explicit definition for logb */
FUNCTION?(logbl, ... long double, ld, ... mpfr, get_ld_2exp)
#endif

FUNCTION1(carg, double complex, double, d, mpc, d, mpfr, arg)
FUNCTION1(cargf, float complex, float, f, mpc, flt, mpfr, arg)
FUNCTION1(csqrt, double complex, double complex, d, mpc, d, mpc, sqrt)
FUNCTION1(csqrtf, float complex, float complex, f, mpc, f, mpc, sqrt)

#endif
#if __FreeBSD_version < 800012

#if 0
/* Need a new macro or explicit definitions for lrtinl and llrintl */
FUNCTION?(llrintl, ... long double, ld, ... mpfr, rint)
FUNCTION?lrintl, ... long double, ld, ... mprfr, rint)
#endif

/* It is sometimes expedient, but not strictly correct, 
 * to conflate rint and nearbyint: they should be 
 * distinguished by the handling of the inexact flag.
 */
FUNCTION1(nearbyintl, long double, long double, ld, mpfr, ld, mpfr, rint)
FUNCTION1(rintl, long double, long double, ld, mpfr, ld, mpfr, rint)
FUNCTION1(exp2l, long double, long double, ld, mpfr, ld, mpfr, exp2)

#endif
#if __FreeBSD_version < 800022

FUNCTION1(sinl, long double, long double, ld, mpfr, ld, mpfr, sin)
FUNCTION1(cosl, long double, long double, ld, mpfr, ld, mpfr, cos)
FUNCTION1(tanl, long double, long double, ld, mpfr, ld, mpfr, tan)
FUNCTION1(tgammaf, float, float, flt, mpfr, flt, mpfr, gamma)

#endif
#if __FreeBSD_version < 800025

FUNCTION1(sqrtl, long double, long double, ld, mpfr, ld, mpfr, sqrt)

#endif
#if __FreeBSD_version < 800030

FUNCTION2(hypotl, long double, ld, mpfr, hypot)
FUNCTION2(remainderl, long double, ld, mpfr, remainder)

#if 0
/* Need a new macro or an explicit definition for remquo */
FUNCTION?(remquol, ... long double, ld, ... mpfr, remquo)
#endif

FUNCTION1(cabsl, long double complex, long double, ld, mpc, ld, mpfr, abs)
FUNCTION1(csqrtl, long double complex, long double complex, ld, mpc, ld, mpc, sqrt)

#endif
#if __FreeBSD_version < 800040

FUNCTION2(fmodl, long double, ld, mpfr, fmod)

#endif
#if __FreeBSD_version < 800042

FUNCTION1(acosl, long double, long double, ld, mpfr, ld, mpfr, acos)
FUNCTION1(asinl, long double, long double, ld, mpfr, ld, mpfr, asin)
FUNCTION1(atanl, long double, long double, ld, mpfr, ld, mpfr, atan)

FUNCTION2(atan2l, long double, ld, mpfr, atan2)

FUNCTION1(cargl, long double complex, long double, ld, mpc, ld, mpfr, arg)
FUNCTION1(cproj, double complex, double complex, d, mpc, d, mpc, proj)
FUNCTION1(cprojf, float complex, float complex, f, mpc, f, mpc, proj)
FUNCTION1(cprojl, long double complex, long double complex, ld, mpc, ld, mpc, proj)

#endif
#if __FreeBSD_version < 900034

FUNCTION1(cexp, double complex, double complex, d, mpc, d, mpc, exp)
FUNCTION1(cexpf, float complex, float complex, f, mpc, f, mpc, exp)

#endif
#if __FreeBSD_version < 900035

FUNCTION1(cbrtl, long double, long double, ld, mpfr, ld, mpfr, cbrt)

#endif
#if __FreeBSD_version < 900045 || \
(__FreeBSD_version >= 1000000 && __FreeBSD_version < 1000001)

FUNCTION1(csin, double complex, double complex, d, mpc, d, mpc, sin)
FUNCTION1(csinf, float complex, float complex, f, mpc, f, mpc, sin)
FUNCTION1(csinh, double complex, double complex, d, mpc, d, mpc, sinh)
FUNCTION1(csinhf, float complex, float complex, f, mpc, f, mpc, sinh)
FUNCTION1(ccos, double complex, double complex, d, mpc, d, mpc, cos)
FUNCTION1(ccosf, float complex, float complex, f, mpc, f, mpc, cos)
FUNCTION1(ccosh, double complex, double complex, d, mpc, d, mpc, cosh)
FUNCTION1(ccoshf, float complex, float complex, f, mpc, f, mpc, cosh)
FUNCTION1(ctan, double complex, double complex, d, mpc, d, mpc, tan)
FUNCTION1(ctanf, float complex, float complex, f, mpc, f, mpc, tan)
FUNCTION1(ctanh, double complex, double complex, d, mpc, d, mpc, tanh)
FUNCTION1(ctanhf, float complex, float complex, f, mpc, f, mpc, tanh)

#endif
#if __FreeBSD_version < 1000016

FUNCTION1(expl, long double, long double, ld, mpfr, ld, mpfr, exp)

#endif

FUNCTION1(acoshl, long double, long double, ld, mpfr, ld, mpfr, acosh)
FUNCTION1(asinhl, long double, long double, ld, mpfr, ld, mpfr, asinh)
FUNCTION1(atanhl, long double, long double, ld, mpfr, ld, mpfr, atanh)

FUNCTION1(cacos, double complex, double complex, d, mpc, d, mpc, acos)
FUNCTION1(cacosf, float complex, float complex, f, mpc, f, mpc, acos)
FUNCTION1(cacosl, long double complex, long double complex, ld, mpc, ld, mpc, acos)

FUNCTION1(cacosh, double complex, double complex, d, mpc, d, mpc, acosh)
FUNCTION1(cacoshf, float complex, float complex, f, mpc, f, mpc, acosh)
FUNCTION1(cacoshl, long double complex, long double complex, ld, mpc, ld, mpc, acosh)

FUNCTION1(casin, double complex, double complex, d, mpc, d, mpc, asin)
FUNCTION1(casinf, float complex, float complex, f, mpc, f, mpc, asin)
FUNCTION1(casinl, long double complex, long double complex, ld, mpc, ld, mpc, asin)

FUNCTION1(casinh, double complex, double complex, d, mpc, d, mpc, asinh)
FUNCTION1(casinhf, float complex, float complex, f, mpc, f, mpc, asinh)
FUNCTION1(casinhl, long double complex, long double complex, ld, mpc, ld, mpc, asinh)

FUNCTION1(catan, double complex, double complex, d, mpc, d, mpc, atan)
FUNCTION1(catanf, float complex, float complex, f, mpc, f, mpc, atan)
FUNCTION1(catanl, long double complex, long double complex, ld, mpc, ld, mpc, atan)

FUNCTION1(catanh, double complex, double complex, d, mpc, d, mpc, atanh)
FUNCTION1(catanhf, float complex, float complex, f, mpc, f, mpc, atanh)
FUNCTION1(catanhl, long double complex, long double complex, ld, mpc, ld, mpc, atanh)

FUNCTION1(ccoshl, long double complex, long double complex, ld, mpc, ld, mpc, cosh)
FUNCTION1(ccosl, long double complex, long double complex, ld, mpc, ld, mpc, cos)
FUNCTION1(cexpl, long double complex, long double complex, ld, mpc, ld, mpc, exp)

FUNCTION1(clog, double complex, double complex, d, mpc, d, mpc, log)
FUNCTION1(clogf, float complex, float complex, f, mpc, f, mpc, log)
FUNCTION1(clogl, long double complex, long double complex, ld, mpc, ld, mpc, log)

FUNCTION1(coshl, long double, long double, ld, mpfr, ld, mpfr, cosh)

FUNCTION1(csinhl, long double complex, long double complex, ld, mpc, ld, mpc, sinh)
FUNCTION1(csinl, long double complex, long double complex, ld, mpc, ld, mpc, sin)
FUNCTION1(ctanhl, long double complex, long double complex, ld, mpc, ld, mpc, tanh)
FUNCTION1(ctanl, long double complex, long double complex, ld, mpc, ld, mpc, tan)

FUNCTION1(erfcl, long double, long double, ld, mpfr, ld, mpfr, erfc)
FUNCTION1(erfl, long double, long double, ld, mpfr, ld, mpfr, erf)
FUNCTION1(expm1l, long double, long double, ld, mpfr, ld, mpfr, expm1)
FUNCTION1(lgammal, long double, long double, ld, mpfr, ld, mpfr, lngamma)
FUNCTION1(log10l, long double, long double, ld, mpfr, ld, mpfr, log10)
FUNCTION1(log1pl, long double, long double, ld, mpfr, ld, mpfr, log1p)
FUNCTION1(log2l, long double, long double, ld, mpfr, ld, mpfr, log2)
FUNCTION1(logl, long double, long double, ld, mpfr, ld, mpfr, log)
FUNCTION1(sinhl, long double, long double, ld, mpfr, ld, mpfr, sinh)
FUNCTION1(tanhl, long double, long double, ld, mpfr, ld, mpfr, tanh)
FUNCTION1(tgammal, long double, long double, ld, mpfr, ld, mpfr, gamma)

FUNCTION2(cpow, double complex, d, mpc, pow)
FUNCTION2(cpowf, float complex, f, mpc, pow)
FUNCTION2(cpowl, long double complex, ld, mpc, pow)

FUNCTION2(powl, long double, ld, mpfr, pow)
